"""
Module to test CLI functionality of click-man package.
"""

import importlib.metadata
import os
from unittest import mock

import click
from click.testing import CliRunner as CLIRunner

from click_man import shell


@mock.patch.object(shell, '_get_entry_point')
def test_cli__missing_entry_point(mock_entry_point):
    mock_entry_point.return_value = None

    runner = CLIRunner()
    result = runner.invoke(shell.cli, 'foo')

    assert result.exit_code == 1, result
    assert 'not an installed console script' in result.output.strip()

    mock_entry_point.assert_called_once_with('foo')


@mock.patch('os.makedirs', new=mock.Mock())
@mock.patch.object(shell, 'write_man_pages')
@mock.patch.object(click, 'echo')
@mock.patch.object(shell, '_get_entry_point')
def test_cli__valid(mock_entry_point, mock_echo, mock_write):
    fake_target = os.path.join(os.getcwd(), 'man')
    fake_command = click.Command(name='foo')
    fake_version = '1.2.3'
    entry_point = mock.Mock(spec=importlib.metadata.EntryPoint)
    entry_point.load.return_value = fake_command
    entry_point.version = fake_version

    mock_entry_point.return_value = entry_point

    runner = CLIRunner()
    result = runner.invoke(shell.cli, ['foo'])

    assert result.exit_code == 0, result.output

    mock_entry_point.assert_called_once_with('foo')
    mock_echo.assert_has_calls(
        [
            mock.call('Load entry point foo'),
            mock.call('Generate man pages for foo in %s' % fake_target),
        ]
    )
    mock_write.assert_called_once_with(
        'foo',
        fake_command,
        version=fake_version,
        target_dir=fake_target,
        date=None,
    )


@mock.patch('os.makedirs', new=mock.Mock())
@mock.patch.object(shell, 'write_man_pages')
@mock.patch.object(click, 'echo')
@mock.patch.object(shell, '_get_entry_point')
def test_cli__with_man_date_version(mock_entry_point, mock_echo, mock_write):
    fake_target = os.path.join(os.getcwd(), 'man')
    fake_command = click.Command(name='foo')
    entry_point = mock.Mock(spec=importlib.metadata.EntryPoint)
    entry_point.load.return_value = fake_command

    mock_entry_point.return_value = entry_point

    runner = CLIRunner()
    result = runner.invoke(
        shell.cli,
        ['foo', '--man-version', '3.2.1', '--man-date', '2020-01-01'],
    )

    assert result.exit_code == 0, result.output

    mock_entry_point.assert_called_once_with('foo')
    mock_echo.assert_has_calls(
        [
            mock.call('Load entry point foo'),
            mock.call('Generate man pages for foo in %s' % fake_target),
        ]
    )
    mock_write.assert_called_once_with(
        'foo',
        fake_command,
        version='3.2.1',
        target_dir=fake_target,
        date='2020-01-01',
    )



"""
Module to test man page functionality of click-man package.
"""

from click_man.man import ManPage


def test_simple_man_page():
    """
    Test creating most simple man page
    """
    man = ManPage('my-command')
    man.short_help = 'Command to test man pages for click.'
    man.date = '21-Feb-1994'
    assert (
        str(man)
        == r""".TH "MY-COMMAND" "1" "21-Feb-1994" "1.0.0" "my-command Manual"
.SH NAME
my-command \- Command to test man pages for click.
.SH SYNOPSIS
.B my-command
"""
    )


def test_full_man_page():
    """
    Test creating man page with all options set
    """
    man = ManPage('my-command')
    man.short_help = 'Command to test man pages for click.'
    man.date = '21-Feb-1994'
    man.synopsis = '[--yolo] [--iambatman]'
    man.description = """This is my awesome

multi line description of a click-man test.

Boaa, richtig geili sach."""

    man.options = [
        ('--yolo', 'Do it in yolo style'),
        ('--iambatman', 'Make me think I am batman'),
    ]
    man.commands = [
        ('start', 'Start it'),
        ('stop', 'Stop it'),
        ('test', 'Pretend you are testing your code'),
    ]

    assert (
        str(man)
        == r""".TH "MY-COMMAND" "1" "21-Feb-1994" "1.0.0" "my-command Manual"
.SH NAME
my-command \- Command to test man pages for click.
.SH SYNOPSIS
.B my-command
[\-\-yolo] [\-\-iambatman]
.SH DESCRIPTION
This is my awesome
.PP
multi line description of a click-man test.
.PP
Boaa, richtig geili sach.
.SH OPTIONS
.TP
\fB\-\-yolo\fP
Do it in yolo style
.TP
\fB\-\-iambatman\fP
Make me think I am batman
.SH COMMANDS
.PP
\fBstart\fP
  Start it
  See \fBmy-command-start(1)\fP for full documentation on the \fBstart\fP command.
.PP
\fBstop\fP
  Stop it
  See \fBmy-command-stop(1)\fP for full documentation on the \fBstop\fP command.
.PP
\fBtest\fP
  Pretend you are testing your code
  See \fBmy-command-test(1)\fP for full documentation on the \fBtest\fP command.
"""
    )  # noqa: E501



"""
click-man - Generate man pages for click application
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides functionality to
write a man page from some given information
about a CLI application.

:copyright: (c) 2016 by Timo Furrer.
:license: MIT, see LICENSE for more details.
"""

import os
import time


class ManPage(object):
    """
    Represent a man page

    :param str command: the name of the command
    """

    TITLE_KEYWORD = '.TH'
    SECTION_HEADING_KEYWORD = '.SH'
    PARAGRAPH_KEYWORD = '.PP'
    BOLD_KEYWORD = '.B'
    INDENT_KEYWORDD = '.TP'

    def __init__(self, command):
        #: Holds the command of the man page
        self.command = command

        #: Holds the short help of the man page
        self.short_help = ''

        #: Holds the synopsis of the man page
        self.synopsis = ''

        #: Holds the description of the man page
        self.description = ''

        #: Holds the version of the man page
        self.version = '1.0.0'

        #: Holds a list of tuple options of the man page
        #  the first item in the tuple are the option switches
        #  and the second one is the option's description
        self.options = []

        #: Holds the commands of the man page
        self.commands = []

        #: Holds the date of the man page creation time.
        self.date = time.strftime(
            '%Y-%m-%d',
            time.gmtime(int(os.environ.get('SOURCE_DATE_EPOCH', time.time()))),
        )

    def replace_blank_lines(self, s):
        """Find any blank lines and replace them with .PP"""

        lines = map(
            lambda x: self.PARAGRAPH_KEYWORD if x == '' else x, s.split('\n')
        )
        return '\n'.join(lines)

    def __str__(self):
        """
        Generate and return the string representation
        of this man page.
        """
        lines = []

        # write title and footer
        lines.append(
            '{0} "{1}" "1" "{2}" "{3}" "{4} Manual"'.format(
                self.TITLE_KEYWORD,
                self.command.upper(),
                self.date,
                self.version,
                self.command,
            )
        )

        # write name section
        lines.append('{0} NAME'.format(self.SECTION_HEADING_KEYWORD))
        lines.append(
            r'{0} \- {1}'.format(
                self.command.replace(' ', r'\-'), self.short_help
            )
        )

        # write synopsis
        lines.append('{0} SYNOPSIS'.format(self.SECTION_HEADING_KEYWORD))
        lines.append('{0} {1}'.format(self.BOLD_KEYWORD, self.command))
        lines.append(self.synopsis.replace('-', r'\-'))

        # write the description
        if self.description:
            lines.append(
                '{0} DESCRIPTION'.format(self.SECTION_HEADING_KEYWORD)
            )
            lines.append(self.replace_blank_lines(self.description))

        # write the options
        if self.options:
            lines.append('{0} OPTIONS'.format(self.SECTION_HEADING_KEYWORD))
            for option, description in self.options:
                lines.append(self.INDENT_KEYWORDD)
                option_unpacked = option.replace('-', r'\-').split()
                lines.append(
                    r'\fB{0}\fP{1}'.format(
                        option_unpacked[0],
                        (' ' + ' '.join(option_unpacked[1:]))
                        if len(option_unpacked) > 1
                        else '',
                    )
                )
                lines.append(self.replace_blank_lines(description))

        # write commands
        if self.commands:
            lines.append('{0} COMMANDS'.format(self.SECTION_HEADING_KEYWORD))
            for name, description in self.commands:
                lines.append(self.PARAGRAPH_KEYWORD)
                lines.append(r'\fB{0}\fP'.format(name))
                lines.append('  ' + self.replace_blank_lines(description))
                lines.append(
                    r'  See \fB{0}-{1}(1)\fP for full documentation on the \fB{1}\fP command.'.format(
                        self.command, name
                    )
                )

        man_page = '\n'.join(lines)

        if not man_page.endswith('\n'):
            # Ensure man page ends with newline.
            man_page += '\n'

        return man_page



"""
click-man - Generate man pages for click application
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides a click CLI command to
generate man pages from a click application.

:copyright: (c) 2016 by Timo Furrer.
:license: MIT, see LICENSE for more details.
"""

from datetime import datetime
import importlib.metadata
import os
import sys
from typing import Optional

import click

from click_man.core import write_man_pages


def _get_entry_point(name: str) -> Optional[importlib.metadata.EntryPoint]:
    entry_points = importlib.metadata.entry_points()
    if sys.version_info >= (3, 10):
        console_scripts = entry_points.select(
            group='console_scripts', name=name
        )
    else:
        console_scripts = [
            ep
            for ep in entry_points.get('console_scripts', [])
            if ep.name == name
        ]

    if len(console_scripts) < 1:
        return None

    # Only generate man pages for first console script
    return tuple(console_scripts)[0]


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option(
    '--target',
    '-t',
    default=os.path.join(os.getcwd(), 'man'),
    type=click.Path(file_okay=False, dir_okay=True, resolve_path=True),
    help='Target location for the man pages',
)
@click.option('--man-version', help='Version to use in generated man page(s)')
@click.option('--man-date', help='Date to use in generated man page(s)')
@click.version_option(
    importlib.metadata.version('click-man'), '-V', '--version'
)
@click.argument('name')
def cli(target, name, man_version, man_date):
    """
    Generate man pages for the scripts defined in the ``console_scripts`` entry
    point.

    The cli application is gathered from entry points of installed packages.

    The generated man pages are written to files in the directory given
    by ``--target``.
    """
    entry_point = _get_entry_point(name)
    if not entry_point:
        raise click.ClickException(
            '"{0}" is not an installed console script.'.format(name)
        )

    # create target directory if it does not exist yet
    try:
        os.makedirs(target)
    except OSError:
        pass

    if not man_version:
        man_version = entry_point.version

    if man_date:
        try:
            datetime.strptime(man_date, '%Y-%m-%d')
        except ValueError:
            raise click.ClickException(
                '"{0}" is not a valid date.'.format(man_date)
            )

    click.echo('Load entry point {0}'.format(name))
    cli = entry_point.load()

    # If the entry point isn't a click.Command object, try to find it in the
    # module
    if not isinstance(cli, click.Command):
        from importlib import import_module
        from inspect import getmembers

        if not entry_point.module:
            raise click.ClickException(
                'Could not find module name for "{0}".'.format(name)
            )

        ep_module = import_module(entry_point.module)
        ep_members = getmembers(
            ep_module,
            lambda x: isinstance(x, click.Command),
        )

        if len(ep_members) < 1:
            raise click.ClickException(
                'Could not find click.Command object for "{0}".'.format(name)
            )

        ep_name, cli = ep_members[0]
        click.echo(
            'Found alternate entry point {0} in {1}'.format(ep_name, name)
        )

    click.echo('Generate man pages for {0} in {1}'.format(name, target))
    write_man_pages(
        name,
        cli,
        version=man_version,
        target_dir=target,
        date=man_date,
    )



"""
click-man - Generate man pages for click application
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the core functionality to
generate man pages for entire click applications.

:copyright: (c) 2016 by Timo Furrer.
:license: MIT, see LICENSE for more details.
"""

import os

import click

from .man import ManPage

CLICK_VERSION = tuple(int(x) for x in click.__version__.split('.')[:2])


def get_short_help_str(command, limit=45):
    """
    Gets short help for the command or makes it by shortening the long help string.
    """
    return (
        command.short_help
        or command.help
        and click.utils.make_default_short_help(command.help, limit)
        or ''
    )


def generate_man_page(ctx, version=None, date=None):
    """
    Generate documentation for the given command.

    :param click.Context ctx: the click context for the
        cli application.
    :param str version: The version information to include in the man page.
    :param str date: The date information to include in the man page.

    :returns: the generate man page from the given click Context.
    :rtype: str
    """
    # Create man page with the details from the given context
    man_page = ManPage(ctx.command_path)
    man_page.version = version
    man_page.short_help = get_short_help_str(ctx.command)
    man_page.description = ctx.command.help
    man_page.synopsis = ' '.join(ctx.command.collect_usage_pieces(ctx))
    man_page.options = [
        x.get_help_record(ctx)
        for x in ctx.command.params
        if isinstance(x, click.Option) and not getattr(x, 'hidden', False)
    ]

    if date:
        man_page.date = date

    commands = getattr(ctx.command, 'commands', None)
    if commands:
        man_page.commands = [
            (k, get_short_help_str(v)) for k, v in commands.items()
        ]

    return str(man_page)


def write_man_pages(
    name,
    cli,
    parent_ctx=None,
    version=None,
    target_dir=None,
    date=None,
):
    """
    Generate man page files recursively
    for the given click cli function.

    :param str name: the cli name
    :param cli: the cli instance
    :param click.Context parent_ctx: the parent click context
    :param str target_dir: the directory where the generated
                           man pages are stored.
    :param date: the date to include in the header
    """
    ctx = click.Context(cli, info_name=name, parent=parent_ctx)

    man_page = generate_man_page(ctx, version)
    path = '{0}.1'.format(ctx.command_path.replace(' ', '-'))
    if target_dir:
        path = os.path.join(target_dir, path)

    with open(path, 'w+') as f:
        f.write(man_page)

    commands = getattr(cli, 'commands', {})
    for name, command in commands.items():
        if CLICK_VERSION >= (7, 0):
            # Since Click 7.0, we have been able to mark commands as hidden
            if command.hidden:
                # Do not write a man page for a hidden command
                continue

        write_man_pages(
            name,
            command,
            parent_ctx=ctx,
            version=version,
            target_dir=target_dir,
            date=date,
        )
